#include <mutex>
#include <iostream>
#include <condition_variable>

using namespace std;

class rwlock {
 public:
  rwlock(): numReaders(0), writeState(Ready) {}
  void acquireAsReader();
  void acquireAsWriter();
  void release();

 private:
  int numReaders;
  enum {
    Ready,   // no one is trying to write
    Pending, // someone is waiting on readers to write
    Writing  // someone is writing
  } writeState;
  mutex readLock, stateLock;
  condition_variable_any readCond, stateCond;
};

void rwlock::acquireAsReader() {
    lock_guard<mutex> lgs(stateLock);
    stateCond.wait(stateLock, [this]{ return writeState == Ready; });
    // why not unlock stateLock here?
    lock_guard<mutex> lgr(readLock);
    numReaders++;
}
 
void rwlock::acquireAsWriter() {
    stateLock.lock();
    stateCond.wait(stateLock, [this]{ return writeState == Ready; });
    writeState = Pending;
    stateLock.unlock();
    // why did we have to unlock stateLock here?
    lock_guard<mutex> lgr(readLock);
    readCond.wait(readLock, [this]{ return numReaders == 0; });
    writeState = Writing;
}
 
void rwlock::release() {
    stateLock.lock();
    if (writeState == Writing) {
        writeState = Ready;
        stateLock.unlock();
        stateCond.notify_all();
        return;
    }
 
    stateLock.unlock();
    lock_guard<mutex> lgr(readLock);
    numReaders--;
    if (numReaders == 0) readCond.notify_one();
}

int main()
{
    std::cout << "Your rwlock compiled!" << std::endl;
}
